"""
MDFEND: Multi-domain Fake News Detection
paper: https://dl.acm.org/doi/10.1145/3459637.3482139
code: https://github.com/kennqiang/MDFEND-Weibo21
"""
from typing import List, Optional, Tuple

import torch
from torch import Tensor
from torch import nn
from transformers import BertModel

from faknow.model.layers.layer import TextCNNLayer
from faknow.model.model import AbstractModel


class _MLP(nn.Module):
    def __init__(self,
                 input_dim: int,
                 embed_dims: List[int],
                 dropout_rate: float,
                 output_layer=True):
        super().__init__()
        layers = list()
        for embed_dim in embed_dims:
            layers.append(nn.Linear(input_dim, embed_dim))
            layers.append(nn.BatchNorm1d(embed_dim))
            layers.append(nn.ReLU())
            layers.append(nn.Dropout(p=dropout_rate))
            input_dim = embed_dim
        if output_layer:
            layers.append(torch.nn.Linear(input_dim, 1))
        self.mlp = torch.nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        """

        Args:
            x (Tensor): shared feature from domain and text, shape=(batch_size, embed_dim)

        """
        return self.mlp(x)


class _MaskAttentionLayer(torch.nn.Module):
    """
    Compute attention layer
    """

    def __init__(self, input_size: int):
        super(_MaskAttentionLayer, self).__init__()
        self.attention_layer = torch.nn.Linear(input_size, 1)

    def forward(self,
                inputs: Tensor,
                mask: Optional[Tensor] = None) -> Tuple[Tensor, Tensor]:
        weights = self.attention_layer(inputs).view(-1, inputs.size(1))
        if mask is not None:
            weights = weights.masked_fill(mask == 0, float("-inf"))
        weights = torch.softmax(weights, dim=-1).unsqueeze(1)
        outputs = torch.matmul(weights, inputs).squeeze(1)
        return outputs, weights


class MDFEND(AbstractModel):
    r"""MDFEND: Multi-domain Fake News Detection
    """

    def __init__(self,
                 pre_trained_bert_name: str,
                 domain_num: int,
                 mlp_dims: Optional[List[int]] = None,
                 dropout_rate=0.2,
                 expert_num=5):
        """

        Args:
            pre_trained_bert_name (str): the name or local path of pre-trained bert model
            domain_num (int): total number of all domains
            mlp_dims (List[int]): a list of the dimensions in MLP layer. Default=[384]
            dropout_rate (float): rate of Dropout layer. Default=0.2
            expert_num (int): number of experts also called TextCNNLayer. Default=5
        """
        super(MDFEND, self).__init__()
        self.domain_num = domain_num
        self.expert_num = expert_num
        self.bert = BertModel.from_pretrained(
            pre_trained_bert_name).requires_grad_(False)
        self.embedding_size = self.bert.config.hidden_size
        self.loss_func = nn.BCELoss()
        if mlp_dims is None:
            mlp_dims = [384]

        filter_num = 64
        filter_sizes = [1, 2, 3, 5, 10]
        experts = [
            TextCNNLayer(self.embedding_size, filter_num, filter_sizes)
            for _ in range(self.expert_num)
        ]
        self.experts = nn.ModuleList(experts)

        self.gate = nn.Sequential(
            nn.Linear(self.embedding_size * 2, mlp_dims[-1]), nn.ReLU(),
            nn.Linear(mlp_dims[-1], self.expert_num), nn.Softmax(dim=1))

        self.attention = _MaskAttentionLayer(self.embedding_size)

        self.domain_embedder = nn.Embedding(num_embeddings=self.domain_num,
                                            embedding_dim=self.embedding_size)
        self.classifier = _MLP(320, mlp_dims, dropout_rate)

    def forward(self, token_id: Tensor, mask: Tensor, domain: Tensor) -> Tensor:
        """

        Args:
            token_id (Tensor): token ids from bert tokenizer, shape=(batch_size, max_len)
            mask (Tensor): mask from bert tokenizer, shape=(batch_size, max_len)
            domain (Tensor): domain id, shape=(batch_size,)

        Returns:
            FloatTensor: the prediction of being fake, shape=(batch_size,)
        """
        text_embedding = self.bert(token_id, attention_mask=mask).last_hidden_state
        attention_feature, _ = self.attention(text_embedding, mask)

        domain_embedding = self.domain_embedder(domain.view(-1, 1)).squeeze(1)

        gate_input = torch.cat([domain_embedding, attention_feature], dim=-1)
        gate_output = self.gate(gate_input)

        shared_feature = 0
        for i in range(self.expert_num):
            expert_feature = self.experts[i](text_embedding)
            shared_feature += (expert_feature * gate_output[:, i].unsqueeze(1))

        label_pred = self.classifier(shared_feature)

        return torch.sigmoid(label_pred.squeeze(1))

    def calculate_loss(self, data) -> Tensor:
        token_ids = data['text']['token_id']
        masks = data['text']['mask']
        domains = data['domain']
        labels = data['label']
        output = self.forward(token_ids, masks, domains)
        return self.loss_func(output, labels.float())

    def predict(self, data_without_label) -> Tensor:
        token_ids = data_without_label['text']['token_id']
        masks = data_without_label['text']['mask']
        domains = data_without_label['domain']

        # shape=(n,), data = 1 or 0
        round_pred = torch.round(self.forward(token_ids, masks,
                                              domains)).long()
        # after one hot: shape=(n,2), data = [0,1] or [1,0]
        one_hot_pred = torch.nn.functional.one_hot(round_pred, num_classes=2)
        return one_hot_pred



from typing import List, Dict

import torch
import yaml
from torch.utils.data import DataLoader
from transformers import BertTokenizer

from faknow.data.dataset.text import TextDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.content_based.mdfend import MDFEND
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import dict2str

__all__ = ['TokenizerMDFEND', 'run_mdfend', 'run_mdfend_from_yaml']


class TokenizerMDFEND:
    def __init__(self, max_len=170, bert="hfl/chinese-roberta-wwm-ext"):
        self.max_len = max_len
        self.tokenizer = BertTokenizer.from_pretrained(bert)

    def __call__(self, texts: List[str]) -> Dict[str, torch.Tensor]:
        inputs = self.tokenizer(texts,
                                return_tensors='pt',
                                max_length=self.max_len,
                                add_special_tokens=True,
                                padding='max_length',
                                truncation=True)
        return {'token_id': inputs['input_ids'], 'mask': inputs['attention_mask']}


def run_mdfend(train_path: str,
               bert='hfl/chinese-roberta-wwm-ext',
               max_len=170,
               domain_num=9,
               batch_size=64,
               num_epochs=50,
               lr=0.0005,
               weight_decay=5e-5,
               step_size=100,
               gamma=0.98,
               metrics: List = None,
               validate_path: str = None,
               test_path: str = None,
               device='cpu'):
    tokenizer = TokenizerMDFEND(max_len, bert)
    train_set = TextDataset(train_path, ['text'], tokenizer)
    train_loader = DataLoader(train_set, batch_size, shuffle=True)

    if validate_path is not None:
        validate_set = TextDataset(validate_path, ['text'], tokenizer)
        val_loader = DataLoader(validate_set, batch_size, shuffle=False)
    else:
        val_loader = None

    model = MDFEND(bert, domain_num)

    optimizer = torch.optim.Adam(params=model.parameters(),
                                 lr=lr,
                                 weight_decay=weight_decay)
    scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size, gamma)
    evaluator = Evaluator(metrics)

    trainer = BaseTrainer(model, evaluator, optimizer, scheduler, device=device)
    trainer.fit(train_loader, num_epochs, validate_loader=val_loader)

    if test_path is not None:
        test_set = TextDataset(test_path, ['text'], tokenizer)
        test_loader = DataLoader(test_set, batch_size, shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print('test result: ', dict2str(test_result))


def run_mdfend_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_mdfend(_config)



from typing import Callable, Optional, List

import torch
import torch.nn.functional as F
from torch import nn, Tensor

from faknow.model.layers.layer import TextCNNLayer
from faknow.model.model import AbstractModel

"""
Convolutional Neural Networks for Sentence Classification
paper: https://aclanthology.org/D14-1181/
code: https://github.com/yoonkim/CNN_sentence
"""


class TextCNN(AbstractModel):
    r"""TextCNN model with a TextCNN layer and a fully connected layer

        Args:
            word_vectors (torch.Tensor): weights of word embedding layer, shape=(vocab_size, embedding_size)
            filter_num (int): number of filters in conv layer. Default=100
            kernel_sizes (List[int]): list of different kernel_num sizes for TextCNNLayer. Default=[3, 4, 5]
            activate_func (Callable): activate function for TextCNNLayer. Default=relu
            dropout (float): drop out rate of fully connected layer. Default=0.5
            freeze (bool): whether to freeze weights in word embedding layer while training. Default=False
        """

    def __init__(self,
                 word_vectors: torch.Tensor,
                 filter_num=100,
                 kernel_sizes: List[int] = None,
                 activate_func: Optional[Callable] = F.relu,
                 dropout=0.5,
                 freeze=False):
        super().__init__()
        if kernel_sizes is None:
            kernel_sizes = [3, 4, 5]

        self.loss_func = nn.CrossEntropyLoss()

        self.word_embedding = nn.Embedding.from_pretrained(word_vectors, freeze=freeze)

        self.text_ccn_layer = TextCNNLayer(word_vectors.shape[-1], filter_num,
                                           kernel_sizes, activate_func)
        self.classifier = nn.Sequential(
            nn.Dropout(dropout),
            nn.Linear(len(kernel_sizes) * filter_num, 2)
        )

    def forward(self, text: torch.Tensor) -> Tensor:
        text = self.word_embedding(text)
        text = self.text_ccn_layer(text)
        out = self.classifier(text)
        return out

    def calculate_loss(self, data) -> Tensor:
        text, label = data
        out = self.forward(text)
        loss = self.loss_func(out, label)
        return loss

    def predict(self, data_without_label):
        if type(data_without_label) is tuple:
            text = data_without_label[0]
        else:
            text = data_without_label
        out = self.forward(text)
        return F.softmax(out, dim=-1)



from typing import List

import torch
import yaml
from torch_geometric.datasets import UPFD
from torch_geometric.loader import DataLoader
from torch_geometric.transforms import ToUndirected

from faknow.evaluate.evaluator import Evaluator
from faknow.model.social_context.gcnfn import GCNFN
from faknow.train.base_gnn_trainer import BaseGNNTrainer

__all__ = ['run_gcnfn', 'run_gcnfn_from_yaml']


def run_gcnfn(root: str,
              name: str,
              feature: str,
              splits=None,
              batch_size=128,
              epochs=110,
              hidden_size=128,
              lr=0.001,
              weight_decay=0.01,
              metrics: List = None,
              device='cpu'):
    if splits is None:
        splits = ['train', 'val', 'test']

    train_dataset = UPFD(root, name, feature, 'train', ToUndirected())
    train_loader = DataLoader(train_dataset,
                              batch_size=batch_size,
                              shuffle=True)

    if 'val' in splits:
        val_dataset = UPFD(root, name, feature, 'val', ToUndirected())
        val_loader = DataLoader(val_dataset,
                                batch_size=batch_size,
                                shuffle=False)
    else:
        val_loader = None

    model = GCNFN(train_dataset.num_features, hidden_size)
    optimizer = torch.optim.Adam(model.parameters(), lr, weight_decay=weight_decay)
    evaluator = Evaluator(metrics)

    trainer = BaseGNNTrainer(model, evaluator, optimizer, device=device)
    trainer.fit(train_loader, epochs, val_loader)

    if 'test' in splits:
        test_dataset = UPFD(root, name, feature, 'test', ToUndirected())
        test_loader = DataLoader(test_dataset,
                                 batch_size=batch_size,
                                 shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(f'test result={test_result}')


def run_gcnfn_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_gcnfn(**_config)



import json
import pickle
from pathlib import Path

import numpy as np
from sklearn.decomposition import TruncatedSVD

ASSETS_PATH = Path(__file__).parent.parent.parent / "assets"
Words = None
We = None
Weight4ind = None


class Params(object):
    def __init__(self):
        self.LW = 1e-5
        self.LC = 1e-5
        self.eta = 0.05

    def __str__(self):
        t = "LW", self.LW, ", LC", self.LC, ", eta", self.eta
        t = map(str, t)
        return ' '.join(t)


params = Params()
params.rmpc = 1


def getWordWeight(weightfile, a=1e-3):
    if a <= 0:  # when the parameter makes no sense, use unweighted
        a = 1.0

    word2weight = {}
    with open(weightfile) as f:
        lines = f.readlines()
    N = 0
    for i in lines:
        i = i.strip()
        if (len(i) > 0):
            i = i.split()
            if (len(i) == 2):
                word2weight[i[0]] = float(i[1])
                N += float(i[1])
            else:
                print(i)
    for key, value in word2weight.items():
        word2weight[key] = a / (a + value / N)
    return word2weight


def getWeight(words, word2weight):
    weight4ind = {}
    for word, ind in words.items():
        if word in word2weight:
            weight4ind[ind] = word2weight[word]
        else:
            weight4ind[ind] = 1.0
    return weight4ind


def lookupIDX(words, w):
    w = w.lower()
    if len(w) > 1 and w[0] == '#':
        w = w.replace("#", "")
    if w in words:
        return words[w]
    elif 'UUUNKKK' in words:
        return words['UUUNKKK']
    else:
        return len(words) - 1


def getSeq(p1, words):
    # p1 = p1.split()
    X1 = []
    for i in p1:
        X1.append(lookupIDX(words, i))
    return X1


def prepare_data(list_of_seqs):
    lengths = [len(s) for s in list_of_seqs]
    n_samples = len(list_of_seqs)
    maxlen = np.max(lengths)
    x = np.zeros((n_samples, maxlen)).astype('int32')
    x_mask = np.zeros((n_samples, maxlen)).astype('float32')
    for idx, s in enumerate(list_of_seqs):
        x[idx, :lengths[idx]] = s
        x_mask[idx, :lengths[idx]] = 1.
    x_mask = np.asarray(x_mask, dtype='float32')
    return x, x_mask


def sentences2idx(sentences, words):
    """
    Given a list of sentences, output array of word indices that can be fed into the algorithms.
    :param sentences: a list of sentences
    :param words: a dictionary, words['str'] is the indices of the word 'str'
    :return: x1, m1. x1[i, :] is the word indices in sentence i, m1[i,:] is the mask for sentence i (0 means no word at the location)
    """
    seq1 = []
    for i in sentences:
        seq1.append(getSeq(i, words))
    x1, m1 = prepare_data(seq1)
    return x1, m1


def seq2weight(seq, mask, weight4ind):
    weight = np.zeros(seq.shape).astype('float32')
    for i in range(seq.shape[0]):
        for j in range(seq.shape[1]):
            if mask[i, j] > 0 and seq[i, j] >= 0:
                weight[i, j] = weight4ind[seq[i, j]]
    weight = np.asarray(weight, dtype='float32')
    return weight


def get_weighted_average(We, x, w):
    """
    Compute the weighted average vectors
    :param We: We[i,:] is the vector for word i
    :param x: x[i, :] are the indices of the words in sentence i
    :param w: w[i, :] are the weights for the words in sentence i
    :return: emb[i, :] are the weighted average vector for sentence i
    """
    n_samples = x.shape[0]
    emb = np.zeros((n_samples, We.shape[1]))
    for i in range(n_samples):
        emb[i, :] = w[i, :].dot(We[x[i, :], :]) / np.count_nonzero(w[i, :])
    return emb


def compute_pc(X, npc=1):
    """
    Compute the principal components. DO NOT MAKE THE DATA ZERO MEAN!
    :param X: X[i,:] is a data point
    :param npc: number of principal components to remove
    :return: component_[i,:] is the i-th pc
    """
    svd = TruncatedSVD(n_components=npc, n_iter=7, random_state=0)
    svd.fit(X)
    return svd.components_


def remove_pc(X, npc=1):
    """
    Remove the projection on the principal components
    :param X: X[i,:] is a data point
    :param npc: number of principal components to remove
    :return: XX[i, :] is the data point after removing its projection
    """
    pc = compute_pc(X, npc)
    if npc == 1:
        XX = X - X.dot(pc.transpose()) * pc
    else:
        XX = X - X.dot(pc.transpose()).dot(pc)
    return XX


def SIF_embedding(We, x, w, params):
    """
    Compute the scores between pairs of sentences using weighted average + removing the projection on the first principal component
    :param We: We[i,:] is the vector for word i
    :param x: x[i, :] are the indices of the words in the i-th sentence
    :param w: w[i, :] are the weights for the words in the i-th sentence
    :param params.rmpc: if >0, remove the projections of the sentence embeddings to their first principal component
    :return: emb, emb[i, :] is the embedding for sentence i
    """
    emb = get_weighted_average(We, x, w)
    if params.rmpc > 0:
        emb = remove_pc(emb, params.rmpc)
    return emb


def sif_embedding(sentences):
    global Words, We, Weight4ind

    if Words is None:
        with open(ASSETS_PATH / "words.json") as f:
            Words = json.load(f)

    if We is None:
        We = np.load(ASSETS_PATH / "We.npy")

    if Weight4ind is None:
        with open(ASSETS_PATH / "weight4ind.pkl", "rb") as f:
            Weight4ind = pickle.load(f)

    x, m = sentences2idx(sentences, Words)
    w = seq2weight(x, m, Weight4ind)
    return SIF_embedding(We, x, w, params)


def preprocess():
    with open(ASSETS_PATH / "words.json") as f:
        Words = json.load(f)

    word2weight = getWordWeight(
        ASSETS_PATH / "enwiki_vocab_min200.txt",
        1e-3,  # the parameter in the SIF weighting scheme, usually in the range [3e-5, 3e-3]
    )
    weight4ind = getWeight(Words, word2weight)

    with open(ASSETS_PATH / "weight4ind.pkl", "wb") as f:
        pickle.dump(weight4ind, f)


def main():
    embd = sif_embedding(
        ["the parameter in the SIF weighting scheme, usually in the range"])
    print(embd.shape)


if __name__ == "__main__":
    # preprocess()
    main()



from torch_geometric.nn import GATConv, GCNConv, SAGEConv

from faknow.model.social_context.base_gnn import _BaseGNN
from faknow.model.social_context.gcnfn import _BaseGCNFN

"""
User Preference-aware Fake News Detection
paper: https://dl.acm.org/doi/abs/10.1145/3404835.3462990
code: https://github.com/safe-graph/GNN-FakeNews
"""


class UPFDGCN(_BaseGNN):
    def __init__(self, feature_size: int, hidden_size=128):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): Default=128
        """
        super().__init__(feature_size, hidden_size, True)
        self.conv = GCNConv(self.feature_size, self.hidden_size)


class UPFDSAGE(_BaseGNN):
    def __init__(self, feature_size: int, hidden_size=128):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): Default=128
        """
        super().__init__(feature_size, hidden_size, True)
        self.conv = SAGEConv(self.feature_size, self.hidden_size)


class UPFDGAT(_BaseGNN):
    def __init__(self, feature_size: int, hidden_size=128):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): Default=128
        """
        super().__init__(feature_size, hidden_size, True)
        self.conv = GATConv(self.feature_size, self.hidden_size)


class UPFDGCNFN(_BaseGCNFN):
    def __init__(self,
                 feature_size: int,
                 hidden_size=128,
                 dropout_ratio=0.5):
        """

            Args:
                feature_size (int): dimension of input node feature
                hidden_size (int): dimension of hidden layer. Default=128
                dropout_ratio (float): dropout ratio. Default=0.5
            """
        super(UPFDGCNFN, self).__init__(feature_size, hidden_size, dropout_ratio, True)



import os
from typing import List, Callable, Any, Dict

import torch

from faknow.data.dataset.text import TextDataset


class MultiModalDataset(TextDataset):
    def __init__(self,
                 path: str,
                 text_features: List[str],
                 tokenize: Callable[[List[str]], Any],
                 image_features: List[str],
                 transform: Callable[[str], Any]):
        super().__init__(path, text_features, tokenize, False)

        self.transform = transform
        self.image_features = []
        for name in image_features:
            self.process_image(name)

        self._to_tensor()

    def __getitem__(self, index) -> Dict[str, Any]:
        item = {}
        for feature_name, feature_values in self.data.items():
            if feature_name in self.image_features:
                value = self.transform(os.path.join(self.root, feature_values[index]))
                if type(value) is not torch.Tensor and type(value) is not dict:
                    raise RuntimeError(
                        'return type of transform function must be tensor')
                # todo 递归字典的情况，是否需要展开
                # self.data.update(new_text)
            elif feature_name in self.text_features and type(
                    feature_values) is dict:
                value = {k: v[index] for k, v in feature_values.items()}
            else:
                value = feature_values[index]
            item[feature_name] = value
        return item

    def process_image(self, name: str):
        self.check_feature(name)
        if name in self.image_features:
            raise ValueError(
                f"'{name}' has already been marked as image features")
        self.image_features.append(name)

    def remove_image(self, name: str):
        self.check_feature(name)
        if name not in self.image_features:
            raise ValueError(f"'{name}' has not been marked as image features")
        self.image_features.remove(name)

    def _to_tensor(self):
        for name, values in self.data.items():
            if name not in self.image_features and name not in self.text_features:
                if type(values[0]) is int or type(values[0]) is float:
                    try:
                        self.data[name] = torch.tensor(values)
                    except RuntimeError as e:
                        e.args = f"fail to convert '{name}' feature into tensor, please check data type"
                        raise e



import pickle
from typing import List, Dict, Any

import torch
import torch.optim
import yaml
from torch import Tensor
from torch.utils.data import TensorDataset, DataLoader

from faknow.evaluate.evaluator import Evaluator
from faknow.model.social_context.eddfn import EDDFN
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import lsh_data_selection

__all__ = ['run_eddfn', 'run_eddfn_from_yaml']


def run_eddfn(train_pool_input: Tensor,
              train_pool_label: Tensor,
              domain_embedding: Tensor,
              budget_size=0.8,
              num_h=10,
              batch_size=32,
              num_epochs=100,
              lr=0.02,
              metrics: List = None,
              device='cpu'):
    input_size = train_pool_input.shape[-1]
    domain_size = domain_embedding.shape[-1]

    train_pool_set = TensorDataset(train_pool_input, domain_embedding, train_pool_label)

    selected_ids = lsh_data_selection(domain_embedding,
                                      int(len(train_pool_set) * budget_size),
                                      num_h)
    train_set = TensorDataset(*train_pool_set[selected_ids])
    train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)

    model = EDDFN(input_size, domain_size)
    evaluator = Evaluator(metrics)
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)

    trainer = BaseTrainer(model, evaluator, optimizer, device=device)
    trainer.fit(train_loader, num_epochs=num_epochs)


def _parse_kargs(config: Dict[str, Any]) -> Dict[str, Any]:
    with open(config['train_pool_input'], 'rb') as f:
        config['train_pool_input'] = pickle.load(f)
    with open(config['train_pool_label'], 'rb') as f:
        config['train_pool_label'] = pickle.load(f)
    with open(config['domain_embedding'], 'rb') as f:
        config['domain_embedding'] = pickle.load(f)
    return config


def run_eddfn_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_eddfn(**_parse_kargs(_config))



from faknow.run.social_context.run_bigcn import run_bigcn_from_yaml, run_bigcn
from faknow.run.social_context.run_eddfn import run_eddfn_from_yaml, run_eddfn
from faknow.run.social_context.run_gcnfn import run_gcnfn_from_yaml, run_gcnfn
from faknow.run.social_context.run_gnncl import run_gnncl_from_yaml, run_gnncl
from faknow.run.social_context.run_nep import run_nep_from_yaml, run_nep
from faknow.run.social_context.run_upfd import run_upfd_from_yaml, run_upfd



from typing import Optional, Callable, Dict, List, Tuple, Any, Set

import torch.utils.data
from torchvision.datasets.folder import find_classes

from faknow.data.dataset.utils import make_dataset, default_walker


class TensorTextDataset(torch.utils.data.Dataset):
    """
    the usage of TensorTextDataset is similar to :class:`torch.utils.data.dataset.TensorDataset`
    """

    def __init__(self,
                 texts: Optional[torch.Tensor] = None,
                 labels: Optional[torch.Tensor] = None,
                 samples: Optional[List[Tuple[torch.Tensor,
                                              torch.Tensor]]] = None,
                 **other_data: torch.Tensor):
        if samples is not None:
            texts = [sample[0] for sample in samples]
            labels = [sample[-1] for sample in samples]
        if len(texts) != len(labels):
            raise RuntimeError(
                'the first dimension between texts and labels must be the same'
            )

        self.texts = texts
        self.labels = labels
        self.other_data = other_data

    def __getitem__(self, index: int) -> Tuple:
        text = self.texts[index]
        label = self.labels[index]
        if self.other_data is None or len(self.other_data) == 0:
            return text, label
        other_data = {k: v[index] for k, v in self.other_data.items()}
        return text, other_data, label

    def __len__(self) -> int:
        return len(self.labels)


class FolderTextDataset(torch.utils.data.Dataset):
    """
    the usage of FolderTextDataset is similar to :class:`torchvision.datasets.ImageFolder`
    the folder should be followings by default: ::
        root/fake/news1.txt
        root/fake/news2.txt
        ...
        root/real/news3.txt
        root/real/news4.txt
    """

    def __init__(
            self,
            root: str,
            embedding: Callable[[str, Optional[Dict]], Any],
            walk_class_dir: Optional[Callable[[str, str, int, List[Tuple], Set],
                                              None]] = default_walker,
            embedding_params: Dict[str, Any] = None,
            **other_data: torch.Tensor):
        self.classes, self.class_to_idx = find_classes(root)
        self.samples = make_dataset(root, self.class_to_idx, None, None,
                                    walk_class_dir)

        self.texts = [sample[0] for sample in self.samples]
        self.labels = [sample[-1] for sample in self.samples]
        self.embedding = embedding
        self.embedding_params = embedding_params
        self.other_data = other_data

    def __getitem__(self, index) -> Tuple:
        text = self.samples[index][0]
        label = self.samples[index][-1]
        if self.embedding_params is None:
            text = self.embedding(text)
        else:
            text = self.embedding(text, self.embedding_params)

        if self.other_data is None or len(self.other_data) == 0:
            return text, label
        other_data = {k: v[index] for k, v in self.other_data.items()}
        return text, other_data, label

    def __len__(self) -> int:
        return len(self.samples)



from typing import Dict, Callable, List, Union

import torch
from torch import Tensor

from faknow.evaluate.metrics import get_metric_func


class Evaluator:
    def __init__(self, metrics: List[Union[str, Callable[[Tensor, Tensor], float]]] = None):
        """generate metric functions for evaluator
        Args:
            metrics: a list of metrics, str or Callable.
                     If the metric is str, built-in metric functions(`accuracy`, `precision`, `recall`, `f1`)
                     will be called according to the metric name.
                     Or the passing metric function with signature `metric_func(outputs: Tensor, y: Tensor) -> float`
                     will be called.
        """
        if metrics is None:
            metrics = ['accuracy', 'precision', 'recall', 'f1']
        self.metrics = {}

        for metric in metrics:
            if type(metric) == str:
                self.metrics[metric] = get_metric_func(metric)
            elif isinstance(metric, Callable):
                self.metrics[metric.__name__] = metric
            else:
                raise RuntimeError(f'only str or callable are supported as metrics, but {type(metric)} are provided')

    def evaluate(self, outputs: torch.Tensor, y: torch.Tensor) -> Dict[str, float]:
        result = {
            metric_name: metric_func(outputs, y)
            for metric_name, metric_func in self.metrics.items()
        }
        return result



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os

from setuptools import setup, find_packages

install_requires = [
    "transformers>=4.26.1",
    "numpy>=1.23.4",
    "pandas>=1.5.2",
    "scikit_learn>=1.1.3",
    "tensorboard>=2.10.0",
    "tqdm>=4.64.1",
    "jieba>=0.42.1",
    "gensim>=4.2.0",
    "pillow>=9.3.0"
]

setup_requires = []

classifiers = ["License :: OSI Approved :: MIT License"]

# Readthedocs requires Sphinx extensions to be specified as part of
# install_requires in order to build properly.
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if on_rtd:
    install_requires.extend(setup_requires)

setup(
    name="faknow",
    version="0.0.1",  # edit faknow/__init__.py in response
    description="A unified fake news detection library",
    url="https://github.com/thinking024/FaKnow",
    author="NPURG",
    author_email="yiyuanzhu@mail.nwpu.edu.cn",
    python_requires=">=3.8.0",
    packages=[package for package in find_packages() if package.startswith("faknow")],
    include_package_data=True,  # include files in MANIFEST.in
    install_requires=install_requires,
    setup_requires=setup_requires,
    zip_safe=False,
    classifiers=classifiers,
)



import datetime
import logging
import os
import sys
from time import time
from typing import Optional, Callable

import torch
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data.dataloader import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from faknow.evaluate.evaluator import Evaluator
from faknow.model.model import AbstractModel
from faknow.utils.util import dict2str


class AbstractTrainer:
    def __init__(
            self,
            model: AbstractModel,
            evaluator: Evaluator,
            optimizer: Optimizer,
            scheduler: Optional[_LRScheduler] = None,
            loss_func: Optional[Callable] = None
    ):
        self.model = model
        self.loss_func = loss_func
        self.optimizer = optimizer
        self.evaluator = evaluator
        self.scheduler = scheduler

    @torch.no_grad()
    def evaluate(
            self,
            data: DataLoader
    ):
        """evaluate after training"""
        raise NotImplementedError

    def fit(
            self,
            train_data: DataLoader,
            num_epochs: int,
            validate_data=None,
            save=False,
            save_path=None
    ):
        raise NotImplementedError


class BaseTrainer(AbstractTrainer):
    def __init__(
            self,
            model: AbstractModel,
            evaluator: Evaluator,
            optimizer: Optimizer,
            scheduler: Optional[_LRScheduler] = None,
            loss_func: Optional[Callable] = None
    ):
        super(BaseTrainer, self).__init__(model, evaluator, optimizer, scheduler, loss_func)

        # create logger
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        """
        # create file handler which logs even debug messages
        fh = logging.FileHandler(logs_path)
        fh.setLevel(logging.DEBUG)
        # create formatter and add it to the handlers
        formatter = logging.Formatter('')
        fh.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(fh)
        """

    @torch.no_grad()
    def evaluate(
            self,
            loader: DataLoader
    ):
        # evaluation mode
        self.model.eval()

        # start evaluating
        outputs = []
        labels = []
        for batch_data in loader:
            outputs.append(self.model.predict(batch_data))
            labels.append(batch_data['label'])
        return self.evaluator.evaluate(torch.concat(outputs), torch.concat(labels))

    def _train_epoch(
            self,
            loader: DataLoader,
            epoch: int,
            writer: SummaryWriter
    ):
        """training for one epoch"""
        # train mode
        self.model.train()

        # initialize tqdm objects
        pbar = tqdm(enumerate(loader), total=len(loader), ncols=100, desc='Training')

        # start training
        loss = others = None
        for batch_id, batch_data in pbar:
            # calculate loss
            result = self.model.calculate_loss(batch_data)

            # check result type
            if type(result) is tuple and len(result) == 2 and type(result[1]) is dict:
                loss = result[0]
                others = result[1]
            elif type(result) is torch.Tensor:
                loss = result
            else:
                raise TypeError(f"result type error: {type(result)}")

            # backward
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            # set progress bar postfix
            pbar.set_postfix_str(f"loss={loss.item():.18f}")

        # close tqdm objects
        pbar.close()

        # train visualization(tensorboard + log + console)
        writer.add_scalar("Train/loss", loss.item(), epoch)
        if others is None:
            self.logger.info(f"training loss : loss={loss.item():.8f}")
            print(f"training loss : loss={loss.item():.8f}", file=sys.stderr)
        else:
            for metric, value in others.items():
                writer.add_scalar("Train/" + metric, value, epoch)
            self.logger.info(f"training loss : loss={loss.item():.8f}    " + dict2str(others))
            print(f"training loss : loss={loss.item():.8f}    " + dict2str(others), file=sys.stderr)

    def _validate_epoch(
            self,
            loader: DataLoader,
            epoch: int,
            writer: SummaryWriter
    ):
        """validation after training for one epoch"""
        # todo 计算best score，作为最佳结果保存
        # evaluate
        result = self.evaluate(loader)

        # validation visualization(tensorboard + log + console)
        for metric, value in result.items():
            writer.add_scalar("Validation/" + metric, value, epoch)
        self.logger.info("validation result : " + dict2str(result))
        print("validation result : " + dict2str(result), file=sys.stderr)

    def save(
            self,
            save_path: str
    ):
        """save the model"""
        if save_path is None:
            save_dir = os.path.join(os.getcwd(), "save")
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            file_name = f"{self.model.__class__.__name__}-{datetime.datetime.now().strftime('%Y-%m-%d-%H_%M_%S')}.pth"
            save_path = os.path.join(save_dir, file_name)

        torch.save(self.model.state_dict(), save_path)

        # save visualization(log + console)
        self.logger.info(f'\nmodel is saved as {save_path}')
        print(f'\nmodel is saved as {save_path}', file=sys.stderr)

    def fit(
            self,
            train_loader: DataLoader,
            num_epochs: int,
            validate_loader: Optional[DataLoader] = None,
            save=False,
            save_path: Optional[str] = None
    ):
        validation = True
        if validate_loader is None:
            validation = False

        # create files(tb_logs + logs)
        tb_logs_path = f"/root/tf-logs/{self.model.__class__.__name__}-{datetime.datetime.now().strftime('%Y-%m-%d-%H_%M_%S')}"
        writer = SummaryWriter(tb_logs_path)

        logs_dir = os.path.join(os.getcwd(), "logs")
        if not os.path.exists(logs_dir):
            os.makedirs(logs_dir)
        file_name = f"{self.model.__class__.__name__}-{datetime.datetime.now().strftime('%Y-%m-%d-%H_%M_%S')}.log"
        logs_path = os.path.join(logs_dir, file_name)
        # create file handler which logs even debug messages
        fh = logging.FileHandler(logs_path)
        fh.setLevel(logging.DEBUG)
        # create formatter and add it to the handlers
        formatter = logging.Formatter('')
        fh.setFormatter(formatter)
        # add the handlers to the logger
        self.logger.addHandler(fh)

        # print some information
        print(f'training data size={len(train_loader.dataset)}', file=sys.stderr)
        self.logger.info(f'training data size={len(train_loader.dataset)}')
        if validation:
            print(f'validation data size={len(validate_loader.dataset)}', file=sys.stderr)
            self.logger.info(f'validation data size={len(validate_loader.dataset)}')
        self.logger.info(f'Tensorboard log is saved as {tb_logs_path}')

        # training for num_epochs
        print('----start training-----', file=sys.stderr)
        for epoch in range(num_epochs):
            print(f'\n--epoch=[{epoch + 1}/{num_epochs}]--', file=sys.stderr)

            # create formatter and add it to the handlers
            formatter = logging.Formatter('')
            fh.setFormatter(formatter)

            # add the handlers to the logger
            self.logger.addHandler(fh)
            self.logger.info(f'\n--epoch=[{epoch + 1}/{num_epochs}]--')

            # create formatter and add it to the handlers
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
            fh.setFormatter(formatter)

            # add the handlers to the logger
            self.logger.addHandler(fh)

            # train
            training_start_time = time()
            self._train_epoch(train_loader, epoch, writer)
            training_end_time = time()
            training_time = training_end_time - training_start_time
            if training_time < 60:
                self.logger.info(f'training time={training_time:.1f}s')
                print(f'training time={training_time:.1f}s', file=sys.stderr)
            else:
                training_time = int(training_time)
                minutes = training_time // 60
                seconds = training_time % 60
                self.logger.info(f'training time={minutes}m{seconds:02d}s')
                print(f'training time={minutes}m{seconds:02d}s', file=sys.stderr)

            # validate
            if validation:
                self._validate_epoch(validate_loader, epoch, writer)

            # learning rate scheduler
            if self.scheduler is not None:
                self.scheduler.step()

        writer.close()

        # save_best the model
        if save:
            self.save(save_path)



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from transformers import BertModel

from faknow.model.model import AbstractModel

"""
Towards Fine-Grained Reasoning for Fake News Detection
paper: https://aaai.org/papers/05746-towards-fine-grained-reasoning-for-fake-news-detection/
code: https://github.com/Ahren09/FinerFact
"""


def kernel_mus(n_kernels):
    """
    get the mean mus for each gaussian kernel_num. Mu is the middle of each bin
    :param n_kernels: number of kernels (including exact match). first one is exact match
    :return: l_mu, a list of mus.
    """
    l_mu = [1]
    if n_kernels == 1:
        return l_mu

    bin_size = 2.0 / (n_kernels - 1)  # score range from [-1, 1]
    l_mu.append(1 - bin_size / 2)  # mus: middle of the bin
    for i in range(1, n_kernels - 1):
        l_mu.append(l_mu[i] - bin_size)
    return l_mu


def kernel_sigmas(n_kernels, sigma_val):
    assert n_kernels >= 1
    l_sigma = [0.001] + [sigma_val] * (n_kernels - 1)
    return l_sigma


class _BertExtractor(nn.Module):
    def __init__(self, bert_name: str):
        super(_BertExtractor, self).__init__()
        self.bert = BertModel.from_pretrained(bert_name)
        self.bert_hidden_size = self.bert.config.hidden_size
        self.dropout = nn.Dropout(0.1)

    def forward(self, input_ids, attention_mask, token_type_ids):
        outputs = self.bert(input_ids=input_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids)
        pooled_output = outputs.pooler_output
        pooled_output = self.dropout(pooled_output)
        last_hidden_state = outputs.last_hidden_state
        last_hidden_state = self.dropout(last_hidden_state)
        return last_hidden_state, pooled_output


class FinerFact(AbstractModel):
    """
    Towards Fine-Grained Reasoning for Fake News Detection
    """

    def __init__(self,
                 bert: str,
                 evidence_num=5,
                 sigma=0.1,
                 kernel_num=11,
                 dropout=0.6,
                 mode='FF',
                 mean=0.0,
                 std=0.001,
                 channel=6,
                 tweet_num=6,
                 user_num=32,
                 word_num=7,
                 user_embed_dim=64,
                 user_embedding: Optional[Tensor] = None,
                 user_field_num=8):
        """

        Args:
            bert (str): name or local path of the bert model
            evidence_num (int): number of evidences. Default=5
            sigma (float): sigma value for gaussian kernel. Default=0.1
            kernel_num (int): number of kernels. Default=11
            dropout (float): dropout rate. Default=0.6
            mode (str): 'FF' or 'FF+'. Default=FF
            mean (float): mean for normal initial weights. Default=0.0
            std (float): std for normal initial weights. Default=0.001
            channel (int): number of channels in each pair. Default=6
            tweet_num (int): number of tweets in each pair. Default=6
            user_num (int): number of users in each pair. Default=32
            word_num (int): number of words in each tweet. Default=7
            user_embed_dim (int): dimension of user embedding. Default=64
            user_embedding (Tensor): pretrained user embedding. If None, user embedding will be generated from scratch. Default=None
            user_field_num (int): number of user fields for training user embedding from scratch. Default=8
        """
        super(FinerFact, self).__init__()

        self.bert_extractor = _BertExtractor(bert)
        self.dropout = nn.Dropout(dropout)
        self.evidence_num = evidence_num
        self.kernel_num = kernel_num
        self.sigma_value = sigma
        self.mode = mode

        # projection layers
        self.proj_inference_de = nn.Linear(
            self.bert_extractor.bert_hidden_size * 2, 2)
        self.proj_att = nn.Linear(self.kernel_num, 1)
        self.proj_input_de = nn.Linear(self.bert_extractor.bert_hidden_size,
                                       self.bert_extractor.bert_hidden_size)
        self.proj_select = nn.Linear(self.kernel_num, 1)

        # gaussian kernels
        self.mus = torch.FloatTensor(kernel_mus(self.kernel_num)).view(
            1, 1, 1, self.kernel_num)
        self.sigmas = torch.FloatTensor(
            kernel_sigmas(self.kernel_num,
                          self.sigma_value)).view(1, 1, 1, self.kernel_num)

        # projection layers
        self.proj_pred_P = nn.Linear(channel, 1)
        self.att_prior_P = nn.Linear(tweet_num, 1, bias=False)
        self.att_prior_U = nn.Linear(user_num, 1, bias=False)
        self.att_prior_K = nn.Linear(word_num, 1, bias=False)
        self.proj_pred_interact = nn.Linear(2, 1)

        # user features
        self.user_embed_dim = user_embed_dim
        self.user_num = user_num

        # whether to use pretrained user embedding
        if user_embedding is None:
            self.user_embedding = nn.Linear(user_field_num,
                                            self.user_embed_dim,
                                            bias=False)
            nn.init.normal_(self.user_embedding.weight, mean, std)
            self.pretrain_user = False
        else:
            self.pretrain_user = True
            self.user_embedding = nn.Embedding.from_pretrained(
                user_embedding.reshape(-1, self.user_embed_dim))

        # gat
        self.proj_gat = nn.Sequential(
            nn.Linear(self.bert_extractor.bert_hidden_size * 2, 128),
            nn.ReLU(True), nn.Linear(128, 1))
        if self.mode == "FF-concat":
            """
            This part has been changed for concatenating
            Ignore this part for now
            """
            self.proj_gat_user = nn.Sequential(
                nn.Linear(self.bert_extractor.bert_hidden_size * 2 +
                          self.user_embed_dim,
                          128,
                          bias=False), nn.ReLU(True),
                nn.Linear(128, 1, bias=False))
        else:
            self.proj_gat_user = nn.Sequential(
                nn.Linear(self.user_embed_dim, 128, bias=False), nn.ReLU(True),
                nn.Linear(128, 1, bias=False))
            self.proj_user = nn.Linear(self.user_embed_dim * 2,
                                       self.bert_extractor.bert_hidden_size *
                                       2,
                                       bias=False)
            nn.init.normal_(self.proj_user.weight, mean, std)

        self.step = -1

        self.__init_weights__(mean, std)

    def __init_weights__(self, mean, std):
        nn.init.normal_(self.att_prior_P.weight, mean, std)
        nn.init.normal_(self.att_prior_U.weight, mean, std)
        nn.init.normal_(self.att_prior_K.weight, mean, std)
        nn.init.normal_(self.proj_gat_user[0].weight, mean, std)
        nn.init.normal_(self.proj_gat_user[2].weight, mean, std)

    def get_user_embed(self, user_metadata, batch_size):
        # todo
        # 使用预训练时，是按batch索引的下标
        # 在训练时传入的是user_embeds_train，而在测试时传入的是user_embeds_test
        # 在训练和测试过程中需要切换
        if self.pretrain_user:
            # select pre-trained user embedding for the current batch
            user_num_batch = self.evidence_num * self.user_num * batch_size
            indices = torch.arange(self.step * user_num_batch,
                                   (self.step + 1) * user_num_batch, 1)
            user_embed = self.user_embedding(indices)
            user_embed = user_embed.reshape(-1, self.evidence_num,
                                            self.user_num, self.user_embed_dim)
        else:
            # generate user embedding from user metadata
            user_embed = self.user_embedding(user_metadata)

        return user_embed

    def att_prior_mr(self, R_p, R_u, R_k):
        # saliency
        H_p = self.att_prior_P(R_p)  # (B, 5, 6)
        H_u = self.att_prior_U(R_u)  # (B, 5, 32)
        H_k = self.att_prior_K(R_k)  # (B, 5, 7)
        delta = H_p + H_u + H_k
        return delta

    def self_attention_user(self,
                            inputs,
                            inputs_hiddens,
                            mask,
                            index,
                            z_qv_z_v_all=None):
        """
        Models interactions among user embeddings
        """

        idx = torch.LongTensor([index])
        mask = mask.view([-1, self.evidence_num, self.user_num])

        # Hidden feature of ONLY current node
        # B, num_user, user_embed_dim
        own_hidden = torch.index_select(inputs_hiddens, 1, idx)
        # B, 1, num_user
        own_mask = torch.index_select(mask, 1, idx)

        # B, 1, user_embed_dim
        # x^v of current (one) user
        own_input = torch.index_select(inputs, 1, idx)

        # B, 5, user_embed_dim
        own_input = own_input.repeat(1, self.evidence_num, 1)

        # Hidden feature of ONLY current node
        # B, 5, num_user, user_embed_dim
        own_hidden = own_hidden.repeat(1, self.evidence_num, 1, 1)

        # B, 5, num_user
        own_mask = own_mask.repeat(1, self.evidence_num, 1)

        # B*5, num_user, user_embed_dim
        hiddens_norm = F.normalize(inputs_hiddens, p=2, dim=-1)

        # B*5, num_user, user_embed_dim
        own_norm = F.normalize(own_hidden, p=2, dim=-1)

        # B*5, num_user
        # gamma: Importance of each user

        att_score = self.get_intersect_matrix_att(
            hiddens_norm.view(-1, self.user_num, self.user_embed_dim),
            own_norm.view(-1, self.user_num, self.user_embed_dim),
            own_mask.view(-1, self.user_num), own_mask.view(-1, self.user_num))

        # B, 5, num_user
        # gamma: Importance of each user
        att_score = att_score.view(-1, self.evidence_num, self.user_num, 1)

        # B, 5, user_embed_dim
        # Token-wise weighted average
        denoise_inputs = torch.sum(att_score * inputs_hiddens, 2)

        if self.mode == "FF-concat":
            z_qv_z_v = z_qv_z_v_all[:, index, :, :]
            concat_att_embed = torch.cat([z_qv_z_v, denoise_inputs], dim=1)
        else:
            concat_att_embed = denoise_inputs

        # weight_de = torch.cat([own_input, denoise_inputs], -1)
        weight_de = self.proj_gat_user(concat_att_embed)
        weight_de = F.softmax(weight_de, dim=1)
        outputs_de = (denoise_inputs * weight_de).sum(dim=1)
        return outputs_de

    def self_attention(self,
                       inputs,
                       inputs_hiddens,
                       mask,
                       mask_evidence,
                       index,
                       trans_mat_prior=None):
        idx = torch.LongTensor([index])
        max_len = mask.shape[-1]

        # bert_pooler_out: B, 5, 768
        # inputs_hiddens: B, 5, 130, 768

        # B, 5, 130
        mask = mask.view([-1, self.evidence_num, max_len])
        # B, 5, 130
        mask_evidence = mask_evidence.view([-1, self.evidence_num, max_len])
        # B, 130, 768
        own_hidden = torch.index_select(inputs_hiddens, 1, idx)

        # B, 1, 130
        own_mask = torch.index_select(mask, 1, idx)

        # B, 1, 768
        # z^v of current (one) claim-evi pair
        own_input = torch.index_select(inputs, 1, idx)
        # B, 5, 130, 768
        own_hidden = own_hidden.repeat(1, self.evidence_num, 1, 1)

        # B, 5, 130
        own_mask = own_mask.repeat(1, self.evidence_num, 1)

        # B, 5, 768
        own_input = own_input.repeat(1, self.evidence_num, 1)

        # B*5, 130
        hiddens_norm = F.normalize(inputs_hiddens, p=2, dim=-1)

        # B, 5, 130, 768
        own_norm = F.normalize(own_hidden, p=2, dim=-1)

        # B*5, 130
        # alpha: Importance of each token
        att_score = self.get_intersect_matrix_att(
            hiddens_norm.view(-1, max_len,
                              self.bert_extractor.bert_hidden_size),
            own_norm.view(-1, max_len, self.bert_extractor.bert_hidden_size),
            mask_evidence.view(-1, max_len), own_mask.view(-1, max_len))

        # B, 5, 130, 1
        # alpha: Token-wise weighted average
        att_score = att_score.view(-1, self.evidence_num, max_len, 1)

        # B, 5, 768

        # z^{q -> p}
        denoise_inputs = torch.sum(att_score * inputs_hiddens, 2)

        # B, 5, 1536
        # z^q || z^{p}
        weight_inp = torch.cat([own_input, inputs], -1)

        # B, 5, 1536
        z_q_z_v = weight_inp

        # MLP()
        # B, 5, 1
        weight_inp = self.proj_gat(weight_inp)

        # gamma
        # B, 5, 1
        weight_inp = F.softmax(weight_inp, dim=1)

        outputs = (inputs * weight_inp).sum(dim=1)

        # B, 5, 1536
        # z^p || z^{q -> p}
        # Can be changed into z^p || z^{q}
        if self.mode == "FF-P":
            weight_de = z_q_z_v  # shallow copy
            denoise_inputs = inputs.clone()
        else:
            weight_de = torch.cat([own_input, denoise_inputs], -1)

        z_qv_z_v = weight_de

        # gamma
        weight_de = self.proj_gat(weight_de)
        if trans_mat_prior is not None:
            weight_de = torch.cat([
                weight_de, trans_mat_prior[:, index].reshape(
                    -1, self.evidence_num, 1)
            ],
                dim=2)
            weight_de = self.proj_pred_interact(weight_de)
        weight_de = F.softmax(weight_de, dim=1)

        # \sum {gamma^{q->p} * \hat{z}^{q->p}}
        outputs_de = (denoise_inputs * weight_de).sum(dim=1)
        return outputs, outputs_de, z_qv_z_v

    def get_intersect_matrix(self, q_embed, d_embed, attn_q, attn_d):
        attn_q = attn_q.view(attn_q.size()[0], attn_q.size()[1], 1)
        attn_d = attn_d.view(attn_d.size()[0], 1, attn_d.size()[1], 1)

        sim = torch.bmm(q_embed,
                        torch.transpose(d_embed, 1,
                                        2)).view(q_embed.size()[0],
                                                 q_embed.size()[1],
                                                 d_embed.size()[1], 1)
        pooling_value = torch.exp(
            (-((sim - self.mus) ** 2) / (self.sigmas ** 2) / 2)) * attn_d
        pooling_sum = torch.sum(
            pooling_value,
            2)  # If merge content and social representation here
        log_pooling_sum = torch.log(torch.clamp(pooling_sum,
                                                min=1e-10)) * attn_q

        log_pooling_sum_all = torch.sum(log_pooling_sum,
                                        1) / (torch.sum(attn_q, 1) + 1e-10)

        log_pooling_sum = self.proj_select(log_pooling_sum_all).view([-1, 1])
        return log_pooling_sum, log_pooling_sum_all

    def get_intersect_matrix_att(self, q_embed, d_embed, attn_q, attn_d):
        attn_q = attn_q.view(attn_q.size()[0], attn_q.size()[1])
        attn_d = attn_d.view(attn_d.size()[0], 1, attn_d.size()[1], 1)
        sim = torch.bmm(q_embed,
                        torch.transpose(d_embed, 1,
                                        2)).view(q_embed.size()[0],
                                                 q_embed.size()[1],
                                                 d_embed.size()[1], 1)

        # B, 130, 11
        pooling_value = torch.exp(
            (-((sim - self.mus) ** 2) / (self.sigmas ** 2) / 2)) * attn_d

        # B*5, user_num, 11
        log_pooling_sum = torch.sum(pooling_value, 2)

        # B*5, user_num, 11
        log_pooling_sum = torch.log(torch.clamp(log_pooling_sum, min=1e-10))

        # B*5, user_num, 1
        log_pooling_sum = self.proj_att(log_pooling_sum).squeeze(-1)

        log_pooling_sum = log_pooling_sum.masked_fill_((1 - attn_q).bool(),
                                                       -1e4)
        log_pooling_sum = F.softmax(log_pooling_sum, dim=1)
        return log_pooling_sum

    def predict_prior(self, score):
        prior = self.proj_score(score)
        return prior

    def reshape_input_and_masks(self, bert_hiddens, mask, type_id):
        max_len = mask.shape[-1]

        # B*5, 130
        mask_text = mask.view(-1, max_len).float()

        # First token ([CLS]) set to 0
        # B*5, 1
        mask_text[:, 0] = 0.0

        # Claim part set to 1 (Except first token [CLS])
        mask_claim = (1 - type_id.float()) * mask_text

        # Evidence part set to 1
        mask_evidence = type_id.float() * mask_text

        # z^p or h_p^0, Hidden representation of first token
        inputs_hiddens = bert_hiddens.view(
            -1, max_len, self.bert_extractor.bert_hidden_size)
        inputs_hiddens_norm = F.normalize(inputs_hiddens, p=2, dim=2)

        return mask_text, mask_claim, mask_evidence, inputs_hiddens, inputs_hiddens_norm

    def channel_text(self, bert_hiddens, bert_pool_out, mask, type_id, delta):
        max_len = mask.shape[-1]

        # mask_text: Every sentence starts with 0, followed by
        # 1's at positions where tokens are present
        mask_text, mask_claim, mask_evidence, inputs_hiddens, inputs_hiddens_norm = self.reshape_input_and_masks(
            bert_hiddens, mask, type_id)

        # Evidence Selection P(K^v, G)
        # log_pooling_sum_all: Content signals at several kernels
        log_pooling_sum, log_pooling_sum_all = self.get_intersect_matrix(
            inputs_hiddens_norm, inputs_hiddens_norm, mask_claim,
            mask_evidence)
        log_pooling_sum = log_pooling_sum.view([-1, self.evidence_num, 1])
        if not self.mode == "FF-I":
            log_pooling_sum += delta
        select_prob = F.softmax(log_pooling_sum, dim=1)

        # Claim Label Prediction: P(y|K^p, G)
        inputs = bert_pool_out.view(
            [-1, self.evidence_num, self.bert_extractor.bert_hidden_size])
        inputs_hiddens = inputs_hiddens.view([
            -1, self.evidence_num, max_len,
            self.bert_extractor.bert_hidden_size
        ])

        inputs_att_de = []
        z_qv_z_v_all = []

        for i in range(self.evidence_num):
            outputs, outputs_de, z_qv_z_v = self.self_attention(
                inputs, inputs_hiddens, mask_text, mask_text, i)
            inputs_att_de.append(outputs_de)
            z_qv_z_v_all.append(z_qv_z_v)

        # All z^{v}, same as `bert_pooler_out`
        inputs_att = inputs.view(
            [-1, self.evidence_num, self.bert_extractor.bert_hidden_size])
        inputs_att_de = torch.cat(inputs_att_de, dim=1)
        z_qv_z_v_all = torch.cat(z_qv_z_v_all, dim=1)

        # All z^{q->v}
        inputs_att_de = inputs_att_de.view(
            [-1, self.evidence_num, self.bert_extractor.bert_hidden_size])
        z_qv_z_v_all = z_qv_z_v_all.view([
            -1, self.evidence_num, self.evidence_num,
            self.bert_extractor.bert_hidden_size
        ])

        return select_prob, inputs_att, inputs_att_de, z_qv_z_v_all

    def channel_user(self, user_hiddens, z_qv_z_v_all=None):

        # ------------------------------------------
        # Evidence Selection: P(\hat{K}^p, G)
        # ------------------------------------------

        user_hiddens = user_hiddens.reshape(-1, self.evidence_num,
                                            self.user_num, self.user_embed_dim)

        mask_usr = torch.ones_like(user_hiddens[:, :, :, 0])

        inputs_att_de_usr = []

        user_inputs = torch.mean(user_hiddens, dim=2)

        for i in range(self.evidence_num):
            outputs_de_usr = self.self_attention_user(user_inputs,
                                                      user_hiddens, mask_usr,
                                                      i, z_qv_z_v_all)
            inputs_att_de_usr.append(outputs_de_usr)

        # All x^{v}, same as `bert_pooler_out`
        inputs_att_usr = user_inputs.view(
            [-1, self.evidence_num, self.user_embed_dim])
        inputs_att_de_usr = torch.cat(inputs_att_de_usr, dim=1)

        inputs_att_de_usr = inputs_att_de_usr.view(
            [-1, self.evidence_num, self.user_embed_dim])

        return inputs_att_usr, inputs_att_de_usr

    def forward(self,
                token_id,
                mask,
                type_id,
                post_rank,
                user_rank,
                keyword_rank,
                user_metadata=None):
        """

        Args:
            token_id (Tensor): shape=(batch_size, evidence_num, max_len)
            mask (Tensor): shape=(batch_size, evidence_num, max_len)
            type_id (Tensor): shape=(batch_size, evidence_num, max_len)
            post_rank (Tensor): ranking of posts, shape=(batch_size, evidence_num, tweet_num)
            user_rank (Tensor): ranking of users, shape=(batch_size, evidence_num, user_num)
            keyword_rank (Tensor): ranking of keywords, shape=(batch_size, evidence_num, word_num)
            user_metadata (Tensor): shape=(batch_size, evidence_num, user_num, user_embed_dim). Default=None

        Returns:
            logits (Tensor): predictions of being fake, shape=(batch_size, 2)
        """
        if (not self.pretrain_user) and user_metadata is None:
            raise ValueError(
                "User metadata is required when pretrained user embedding is not provided."
            )
        if self.pretrain_user and user_metadata is not None:
            print(
                "Warning: User metadata won't be used when pretrained user embedding is provided."
            )

        self.step = self.step + 1
        max_len, batch_size = token_id.shape[-1], token_id.shape[0]

        # reshape
        token_id = token_id.view(-1, max_len)
        mask = mask.view(-1, max_len)
        type_id = type_id.view(-1, max_len)

        bert_hidden_states, bert_pool_out = self.bert_extractor(
            token_id, mask, type_id)
        delta = self.att_prior_mr(post_rank, user_rank, keyword_rank)

        select_prob, inputs_att, inputs_att_de, z_qv_z_v_all = self.channel_text(
            bert_hidden_states, bert_pool_out, mask, type_id, delta)
        # All z^{q->v} || All z^{v}
        inputs_att = torch.cat([inputs_att, inputs_att_de], -1)

        # user features
        user_embed = self.get_user_embed(user_metadata, batch_size)

        if self.mode in ["FF"]:
            # Initialize our embedding module from the embedding matrix
            inputs_att_usr, inputs_att_de_usr = self.channel_user(
                user_embed, z_qv_z_v_all)
            inputs_att_usr_combined = torch.cat(
                [inputs_att_usr, inputs_att_de_usr], -1)
            inputs_att_usr_combined = self.proj_user(inputs_att_usr_combined)

            assert inputs_att_usr_combined.shape == inputs_att.shape
            inputs_att = inputs_att + inputs_att_usr_combined

        inference_feature = self.proj_inference_de(inputs_att)
        class_prob = F.softmax(inference_feature, dim=2)

        if self.mode == "FF-I":
            select_prob = torch.ones_like(select_prob) / self.evidence_num

        prob = torch.sum(select_prob * class_prob, 1)
        logits = torch.log(prob)

        return logits

    def calculate_loss(self, data) -> Tensor:
        # token_id, mask, type_id, label, post_rank, user_rank, keyword_rank, user_metadata = data
        """get item from batch dict"""
        token_id = data['token_id']
        mask = data['mask']
        type_id = data['type_id']
        label = data['label']
        post_rank = data['post_rank']
        user_rank = data['user_rank']
        keyword_rank = data['keyword_rank']
        user_metadata = data.get('user_metadata', None)
        logits = self.forward(token_id, mask, type_id, post_rank, user_rank, keyword_rank,
                              user_metadata)
        loss = F.nll_loss(logits, label)
        return loss

    def predict(self, data):
        # token_id, mask, type_id, post_rank, user_rank, keyword_rank, user_metadata = data_without_label
        token_id = data['token_id']
        mask = data['mask']
        type_id = data['type_id']
        post_rank = data['post_rank']
        user_rank = data['user_rank']
        keyword_rank = data['keyword_rank']
        user_metadata = data.get('user_metadata', None)
        logits = self.forward(token_id, mask, type_id, post_rank, user_rank, keyword_rank,
                              user_metadata)
        return logits



from typing import Union, Dict

import torch.nn as nn
from torch import Tensor


class AbstractModel(nn.Module):
    """abstract class for all models"""

    def __init__(self):
        super(AbstractModel, self).__init__()

    def calculate_loss(self, data) -> Union[Tensor, Dict[str, Tensor]]:
        raise NotImplementedError

    def predict(self, data_without_label) -> Tensor:
        raise NotImplementedError

    # todo 如果用val loss进行early stopping
    # 那么需要 一个函数调用forward，既能计算loss，又能predict
    # 这样进行validate时，能看到loss与accuracy



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

"""
layers for transformer, including:
Feed-Forward Networks
AddNorm
Scaled Dot Product Attention
Mutil-head Attention
Encoder Layer
"""


def sequence_mask(x: Tensor, valid_len: Optional[Tensor] = None, value=0.):
    """Mask irrelevant entries in sequences."""
    max_len = x.size(1)
    mask = torch.arange(max_len, dtype=torch.float32,
                        device=x.device)[None, :] < valid_len[:, None]
    x[~mask] = value
    return x


def masked_softmax(x: Tensor, valid_lens: Optional[Tensor] = None):
    """Perform softmax operation by masking elements on the last axis."""
    # x: 3D tensor, valid_lens: 1D or 2D tensor
    if valid_lens is None:
        return nn.functional.softmax(x, dim=-1)
    else:
        shape = x.shape
        if valid_lens.dim() == 1:
            valid_lens = torch.repeat_interleave(valid_lens, shape[1])
        else:
            valid_lens = valid_lens.reshape(-1)
        # On the last axis, replace masked elements with a very large negative
        # value, whose exponentiation outputs 0
        x = sequence_mask(x.reshape(-1, shape[-1]), valid_lens, value=-1e6)
        return nn.functional.softmax(x.reshape(shape), dim=-1)


def transpose_qkv(x, num_heads):
    """
    Transposition for parallel computation of multiple attention heads.
    Args:
        x: (batch_size, num, num_hiddens), num_hiddens = head_num * out_size
        num_heads: number of attention heads

    Returns:
        (batch_size * head_num, num, num_hiddens/head_num)
    """

    # after:
    # (batch_size, num, head_num, num_hiddens/head_num)
    x = x.reshape(x.shape[0], x.shape[1], num_heads, -1)

    # after:
    # (batch_size, head_num, num, num_hiddens/head_num)
    x = x.permute(0, 2, 1, 3)

    return x.reshape(-1, x.shape[2], x.shape[3])


def transpose_output(x, num_heads):
    """
    Reverse the operation of transpose_qkv.
    Args:
        x: (batch_size * head_num, num, num_hiddens/head_num)
        num_heads: number of attention heads

    Returns:
        (batch_size, num, num_hiddens), num_hiddens = head_num * out_size
    """
    x = x.reshape(-1, num_heads, x.shape[1], x.shape[2])
    x = x.permute(0, 2, 1, 3)
    return x.reshape(x.shape[0], x.shape[1], -1)


class FFN(nn.Module):
    def __init__(self, input_size: int,
                 hidden_size: int,
                 output_size: Optional[int] = None,
                 dropout=0.,
                 activation=F.relu):
        super(FFN, self).__init__()

        if output_size is None:
            output_size = input_size

        self.dense1 = nn.Linear(input_size, hidden_size)
        self.activation = activation
        self.dense2 = nn.Linear(hidden_size, output_size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor):
        x = self.dense2(self.activation(self.dense1(x)))
        return self.dropout(x)


class AddNorm(nn.Module):
    def __init__(self, normalized_shape: int, dropout=0.):
        super(AddNorm, self).__init__()
        self.dropout = nn.Dropout(dropout)
        self.ln = nn.LayerNorm(normalized_shape)

    def forward(self, x: Tensor, y: Tensor):
        """
        Args:
            x: residual
            y: output of sublayer
        """
        return self.ln(self.dropout(y) + x)


class ScaledDotProductAttention(nn.Module):
    def __init__(self, dropout=0., epsilon=0.):
        super(ScaledDotProductAttention, self).__init__()
        self.epsilon = epsilon
        self.dropout = nn.Dropout(dropout)
        self.attention_weights = None

    def forward(self,
                queries: Tensor,
                keys: Tensor,
                values: Tensor,
                valid_lens: Optional[Tensor] = None):
        """
        Args:
            queries: (batch_size, q_num, d)
            keys: (batch_size, k-v_num, d)
            values: (batch_size, k-v_num, v_dim)
            valid_lens: (batch_size,) or (batch_size, q_num)

        Returns:
            attention_values: (batch_size, q_num, v_dim)
        """
        d = queries.shape[-1]
        scores = torch.bmm(queries, keys.transpose(
            1, 2)) / (d ** 0.5 + self.epsilon)
        self.attention_weights = masked_softmax(scores, valid_lens)
        return torch.bmm(self.dropout(self.attention_weights), values)


class MultiHeadAttention(nn.Module):
    def __init__(self,
                 input_size: int,
                 k_out_size: int,
                 v_out_size: int,
                 head_num: int,
                 out_size: Optional[int] = None,
                 q_in_size: Optional[int] = None,
                 k_in_size: Optional[int] = None,
                 v_in_size: Optional[int] = None,
                 dropout=0.,
                 bias=False):
        super(MultiHeadAttention, self).__init__()
        self.head_num = head_num
        self.q_in_size = q_in_size if q_in_size is not None else input_size
        self.k_in_size = k_in_size if k_in_size is not None else input_size
        self.v_in_size = v_in_size if v_in_size is not None else input_size
        self.out_size = out_size if out_size is not None else input_size

        self.attention = ScaledDotProductAttention(dropout)
        self.W_q = nn.Linear(self.q_in_size, k_out_size * head_num, bias=bias)
        self.W_k = nn.Linear(self.k_in_size, k_out_size * head_num, bias=bias)
        self.W_v = nn.Linear(self.v_in_size, v_out_size * head_num, bias=bias)
        self.W_o = nn.Linear(v_out_size * head_num, self.out_size * head_num, bias=bias)

    def forward(self, queries, keys, values, valid_lens=None):
        """
        Args:
            queries: (batch_size, q_num, d)
            keys: (batch_size, k-v_num, d)
            values: (batch_size, k-v_num, v-dim)
            valid_lens: (batch_size,) or (batch_size, q_num)

        Returns:
            multi-head output: (batch_size, q_num, out_size * head_num)
        """

        # After transpose:
        # (batch_size * head_num, num, out_size)
        queries = transpose_qkv(self.W_q(queries), self.head_num)
        keys = transpose_qkv(self.W_k(keys), self.head_num)
        values = transpose_qkv(self.W_v(values), self.head_num)

        if valid_lens is not None:
            # On axis 0, copy the first item (scalar or vector) for
            # head_num times, then copy the next item, and so on
            valid_lens = torch.repeat_interleave(valid_lens,
                                                 repeats=self.head_num,
                                                 dim=0)

        # (batch_size * head_num, q_num, v_out_size)
        output = self.attention(queries, keys, values, valid_lens)

        # After transpose:
        # (batch_size, q_num, head_num * v_out_size)
        output_concat = transpose_output(output, self.head_num)
        return self.W_o(output_concat)


class EncoderLayer(nn.Module):
    def __init__(self,
                 input_size: int,
                 ffn_hidden_size: int,
                 head_num: int,
                 k_out_size: int,
                 v_out_size: int,
                 dropout=0.,
                 bias=False):
        super(EncoderLayer, self).__init__()
        self.attention = MultiHeadAttention(input_size,
                                            k_out_size,
                                            v_out_size,
                                            head_num,
                                            dropout=dropout,
                                            bias=bias)
        self.addnorm1 = AddNorm(input_size, dropout)
        self.ffn = FFN(input_size,
                       ffn_hidden_size,
                       input_size,
                       dropout)
        self.addnorm2 = AddNorm(input_size, dropout)

    def forward(self, x, valid_lens=None):
        y = self.addnorm1(x, self.attention(x, x, x, valid_lens))
        return self.addnorm2(y, self.ffn(y))


from typing import Tuple, Dict

import torch
import torch.nn.functional as F
from torch import Tensor
from torch import nn

from faknow.model.layers.transformer import FFN
from faknow.model.model import AbstractModel

"""
Embracing Domain Differences in Fake News Cross-domain Fake News Detection using Multi-modal Data
paper: https://ojs.aaai.org/index.php/AAAI/article/view/16134
code: github.com/amilasilva92/cross-domain-fake-news-detection-aaai2021
"""


class _Discriminator(nn.Module):
    def __init__(self, input_size: int, domain_size: int):
        super().__init__()
        self.ffn = FFN(input_size, domain_size * 2, domain_size, activation=torch.sigmoid)
        self.sigmoid = nn.Sigmoid()

    def forward(self, input_representation: Tensor):
        return self.sigmoid(self.ffn(input_representation))


class EDDFN(AbstractModel):
    r"""Embracing Domain Differences in Fake News Cross-domain Fake News Detection using Multi-modal Data

        Args:
            input_size (int): dimension of input representation
            domain_size (int): dimension of domain vector
            lambda1 (float): L_{recon} loss weight. Default: 1.0
            lambda2 (float): L_{specific} loss weight. Default: 10.0
            lambda3 (float): L_{shared} loss weight. Default: 5.0
            hidden_size (int): size of hidden layer. Default: 512
    """

    def __init__(self, input_size: int, domain_size: int, lambda1=1.0, lambda2=10.0, lambda3=5.0,
                 hidden_size=512):

        super().__init__()
        self.input_size = input_size
        self.domain_size = domain_size
        self.lambda1 = lambda1
        self.lambda2 = lambda2
        self.lambda3 = lambda3
        self.discriminate_size = int(hidden_size / 2)

        self.input_embedding_layer = nn.Sequential(FFN(input_size, self.discriminate_size, hidden_size), nn.ReLU())

        self.output_classifier = nn.Sequential(nn.Linear(hidden_size, 1), nn.Sigmoid())
        self.output_decoder = FFN(hidden_size, self.discriminate_size, input_size)

        self.specific_discriminator = _Discriminator(self.discriminate_size, domain_size)
        self.shared_discriminator = _Discriminator(self.discriminate_size, domain_size)

    def forward(self, input_representation: Tensor) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """

        Args:
            input_representation (Tensor): input representation consisting of text and propagation network

        Returns:
            tuple:
                - class_out (Tensor): prediction of being fake news, shape=(batch_size, 1)
                - decoder_out (Tensor): prediction of input, shape=(batch_size, input_size)
                - specific_domain (Tensor): specific domain output, shape=(batch_size, domain_size)
                - shared_domain (Tensor): shared domain output, shape=(batch_size, domain_size)

        """
        input_embedding = self.input_embedding_layer(input_representation)

        # shape=(batch_size * 1)
        class_out = self.output_classifier(input_embedding)

        decoder_out = self.output_decoder(input_embedding)

        specific_input = input_embedding[:, :self.discriminate_size]
        shared_input = input_embedding[:, self.discriminate_size:]

        specific_domain = self.specific_discriminator(specific_input)
        shared_domain = self.shared_discriminator(shared_input)

        return class_out, decoder_out, specific_domain, shared_domain

    def calculate_loss(self, data) -> Dict[str, Tensor]:
        input_representation, domain, label = data
        class_out, decoder_out, specific_domain, shared_domain = self.forward(input_representation)

        class_loss = nn.BCELoss()(class_out.squeeze(), label.float())
        decoder_loss = nn.MSELoss()(decoder_out, input_representation) * self.lambda1
        specific_domain_loss = nn.MSELoss()(specific_domain, domain) * self.lambda2
        shared_domain_loss = nn.MSELoss()(shared_domain, domain) * self.lambda3

        loss = class_loss + decoder_loss + specific_domain_loss + shared_domain_loss

        return {'total_loss': loss, 'class_loss': class_loss, 'decoder_loss': decoder_loss,
                'specific_domain_loss': specific_domain_loss,
                'shared_domain_loss': shared_domain_loss}

    def predict(self, data_without_label):
        if type(data_without_label) is tuple:
            input_representation = data_without_label[0]
        else:
            input_representation = data_without_label

        class_out = self.forward(input_representation)[0]
        class_out = torch.round(class_out).long().squeeze()
        return F.one_hot(class_out, num_classes=2)



from typing import List, Optional

import torch
import yaml
from torch.utils.data import DataLoader

from faknow.data.dataset.safe_dataset import SAFENumpyDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.content_based.multi_modal.safe import SAFE
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import dict2str

__all__ = ['run_safe', 'run_safe_from_yaml']


def run_safe(
        train_path: str,
        validate_path: str = None,
        test_path: str = None,
        embedding_size: int = 300,
        conv_in_size: int = 32,
        filter_num: int = 128,
        cnn_out_size: int = 200,
        dropout: float = 0.,
        loss_weights: Optional[List[float]] = None,
        batch_size=64,
        lr=0.00025,
        metrics: List = None,
        num_epochs=100,
        device='cpu',
):
    train_dataset = SAFENumpyDataset(train_path)
    train_loader = DataLoader(dataset=train_dataset, batch_size=batch_size, shuffle=True)

    if validate_path is not None:
        val_dataset = SAFENumpyDataset(validate_path)
        val_loader = DataLoader(dataset=val_dataset, batch_size=batch_size, shuffle=True)
    else:
        val_loader = None

    model = SAFE(embedding_size, conv_in_size, filter_num, cnn_out_size, dropout, loss_weights)
    optimizer = torch.optim.Adam(
        filter(
            lambda p: p.requires_grad,
            list(model.parameters())),
        lr
    )
    evaluator = Evaluator(metrics)

    trainer = BaseTrainer(model=model, evaluator=evaluator, optimizer=optimizer, device=device)
    trainer.fit(train_loader=train_loader, num_epochs=num_epochs, validate_loader=val_loader)

    if test_path is not None:
        test_set = SAFENumpyDataset(test_path)
        test_loader = DataLoader(dataset=test_set, batch_size=batch_size, shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(f"test result: {dict2str(test_result)}")


def run_safe_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_safe(**_config)



from typing import List, Dict

import numpy as np
import torch
import torch.nn.functional as F
import yaml
from PIL import Image
from scipy.fftpack import fft, dct
from torch.utils.data import DataLoader
from torchvision.transforms import transforms
from transformers import get_linear_schedule_with_warmup, BertTokenizer

from faknow.data.dataset.multi_modal import MultiModalDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.content_based.multi_modal.mcan import MCAN
from faknow.train.trainer import BaseTrainer

__all__ = ['TokenizerMCAN', 'transform_mcan', 'process_dct_mcan', 'get_optimizer_mcan', 'run_mcan',
           'run_mcan_from_yaml']


class TokenizerMCAN:
    def __init__(self, max_len=160, bert="bert-base-chinese"):
        self.max_len = max_len
        self.tokenizer = BertTokenizer.from_pretrained(bert)

    def __call__(self, texts: List[str]) -> Dict[str, torch.Tensor]:
        inputs = self.tokenizer(texts,
                                return_tensors='pt',
                                max_length=self.max_len,
                                add_special_tokens=True,
                                padding='max_length',
                                truncation=True)
        return {'token_id': inputs['input_ids'], 'mask': inputs['attention_mask']}


def transform_mcan(path: str) -> Dict[str, torch.Tensor]:
    with open(path, "rb") as f:
        img = Image.open(f)
        transform_vgg = transforms.Compose([
            transforms.Resize((224, 224)),
            transforms.ToTensor(),
        ])
        vgg_feature = transform_vgg(img.convert('RGB'))

        transform_dct = transforms.Compose([
            transforms.Resize((224, 224)),
            transforms.ToTensor()
        ])
        dct_feature = process_dct_mcan(transform_dct(img.convert('L')))

    return {'vgg': vgg_feature, 'dct': dct_feature}


def process_dct_mcan(img: torch.Tensor) -> torch.Tensor:
    img = img.numpy()  # size = [1, 224, 224]
    height = img.shape[1]
    width = img.shape[2]
    N = 8
    step = int(height / N)

    dct_img = np.zeros((1, N * N, step * step, 1), dtype=np.float32)
    fft_img = np.zeros((1, N * N, step * step, 1))

    i = 0
    for row in np.arange(0, height, step):
        for col in np.arange(0, width, step):
            block = np.array(img[:, row:(row + step), col:(col + step)], dtype=np.float32)
            block1 = block.reshape((-1, step * step, 1))  # [batch_size,784,1]
            dct_img[:, i, :, :] = dct(block1)  # [batch_size, 64, 784, 1]
            i += 1

    # for i in range(64):
    fft_img[:, :, :, :] = fft(dct_img[:, :, :, :]).real  # [batch_size,64, 784,1]

    fft_img = torch.from_numpy(fft_img).float()  # [batch_size, 64, 784, 1]
    new_img = F.interpolate(fft_img, size=[250, 1])  # [batch_size, 64, 250, 1]
    new_img = new_img.squeeze(0).squeeze(-1)  # torch.size = [64, 250]

    return new_img


def get_optimizer_mcan(model: MCAN,
                       lr=0.0001,
                       weight_decay=0.15,
                       bert_lr=1e-5,
                       vgg_lr=1e-5,
                       dtc_lr=1e-5,
                       fusion_lr=1e-2,
                       linear_lr=1e-2,
                       classifier_lr=1e-2):
    no_decay = [
        "bias",
        "gamma",
        "beta",
        "LayerNorm.weight",
        "bn_text.weight",
        "bn_dct.weight",
        "bn_1.weight",
    ]

    bert_params = list(model.bert.named_parameters())
    vgg_params = list(model.vgg.named_parameters())
    dtc_params = list(model.dct_img.named_parameters())
    fusion_params = list(
        model.fusion_layers.named_parameters()
    )
    linear_params = (
            list(model.linear_text.named_parameters())
            + list(model.linear_vgg.named_parameters())
            + list(model.linear_dct.named_parameters())
    )
    classifier_params = list(model.linear1.named_parameters()) + list(
        model.linear2.named_parameters()
    )

    optimizer_grouped_parameters = [
        # bert_param_optimizer
        {"params": [p for n, p in bert_params if not any(nd in n for nd in no_decay)],
         "weight_decay": weight_decay,
         "lr": bert_lr, },
        {"params": [p for n, p in bert_params if any(nd in n for nd in no_decay)],
         "weight_decay": 0.0,
         "lr": bert_lr, },

        # vgg_param_optimizer
        {"params": [p for n, p in vgg_params if not any(nd in n for nd in no_decay)],
         "weight_decay": weight_decay,
         "lr": vgg_lr, },
        {"params": [p for n, p in vgg_params if any(nd in n for nd in no_decay)],
         "weight_decay": 0.0,
         "lr": vgg_lr, },

        # dtc_conv_param_optimizer
        {"params": [p for n, p in dtc_params if not any(nd in n for nd in no_decay)],
         "weight_decay": weight_decay,
         "lr": dtc_lr, },
        {"params": [p for n, p in dtc_params if any(nd in n for nd in no_decay)],
         "weight_decay": 0.0,
         "lr": dtc_lr, },

        # fusion_param_optimizer
        {"params": [p for n, p in fusion_params if not any(nd in n for nd in no_decay)],
         "weight_decay": weight_decay,
         "lr": fusion_lr, },
        {"params": [p for n, p in fusion_params if any(nd in n for nd in no_decay)],
         "weight_decay": 0.0,
         "lr": fusion_lr, },

        # linear_param_optimizer
        {"params": [p for n, p in linear_params if not any(nd in n for nd in no_decay)],
         "weight_decay": weight_decay,
         "lr": linear_lr, },
        {"params": [p for n, p in linear_params if any(nd in n for nd in no_decay)],
         "weight_decay": 0.0,
         "lr": linear_lr, },

        # classifier_param_optimizer
        {"params": [p for n, p in classifier_params if not any(nd in n for nd in no_decay)],
         "weight_decay": weight_decay,
         "lr": classifier_lr, },
        {"params": [p for n, p in classifier_params if any(nd in n for nd in no_decay)],
         "weight_decay": 0.0,
         "lr": classifier_lr, },
    ]

    optimizer = torch.optim.AdamW(
        optimizer_grouped_parameters,
        lr=lr,
        weight_decay=weight_decay,
    )
    return optimizer


def get_scheduler(batch_num, epoch_num, optimizer, warm_up_percentage=0.1):
    # Total number of training steps is number of batches * number of epochs.
    total_steps = batch_num * epoch_num

    # Create the learning rate scheduler.
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=round(total_steps * warm_up_percentage),
        num_training_steps=total_steps
    )
    return scheduler


def run_mcan(train_path: str,
             bert='bert-base-chinese',
             max_len=255,
             batch_size=16,
             num_epochs=100,
             metrics: List = None,
             validate_path: str = None,
             test_path: str = None,
             device='cpu',
             **optimizer_kargs):
    tokenizer = TokenizerMCAN(max_len, bert)

    train_dataset = MultiModalDataset(train_path, ['text'], tokenizer, ['image'], transform_mcan)
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

    if validate_path:
        validation_dataset = MultiModalDataset(validate_path, ['text'], tokenizer, ['image'], transform_mcan)
        val_loader = DataLoader(validation_dataset, batch_size=batch_size, shuffle=False)
    else:
        val_loader = None

    model = MCAN(bert)
    optimizer = get_optimizer_mcan(model, **optimizer_kargs)
    scheduler = get_scheduler(len(train_loader), num_epochs, optimizer)
    evaluator = Evaluator(metrics)
    clip_grad_norm = {'max_norm': 1.0}

    trainer = BaseTrainer(model, evaluator, optimizer, scheduler, clip_grad_norm, device=device)
    trainer.fit(train_loader, num_epochs=num_epochs, validate_loader=val_loader)

    if test_path:
        test_dataset = MultiModalDataset(test_path, ['text'], tokenizer, ['image'], transform_mcan)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(test_result)


def run_mcan_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_mcan(**_config)



from copy import copy

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch_geometric.nn import GCNConv
from torch_scatter import scatter_mean

from faknow.model.model import AbstractModel

"""
Rumor Detection on Social Media with Bi-Directional Graph Convolutional Networks
paper: https://ojs.aaai.org/index.php/AAAI/article/view/5393
code: https://github.com/safe-graph/GNN-FakeNews
"""


class _RumorGCN(torch.nn.Module):
    def __init__(self,
                 feature_size: int,
                 hidden_size: int,
                 out_size: int,
                 dropout_ratio=0.5):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): dimension of hidden layer
            out_size (int): dimension of output layer
            dropout_ratio (float): drop out rate. Default=0.5
        """
        super(_RumorGCN, self).__init__()
        self.dropout_ratio = dropout_ratio
        self.conv1 = GCNConv(feature_size, hidden_size)
        self.conv2 = GCNConv(hidden_size + feature_size, out_size)

    def forward(self, x: Tensor, edge_index: Tensor, batch: Tensor,
                root_index: Tensor):
        """

        Args:
            x (Tensor): node feature, shape=(num_nodes, feature_size)
            edge_index (Tensor): edge index, shape=(2, num_edges)
            batch (Tensor): index of graph each node belongs to, shape=(num_nodes,)
            root_index (Tensor): index of root node for each graph, shape=(num_graphs,)

        Returns:
            out (Tensor): output of graph convolution layer, shape=(batch_size, 2 * out_size)
        """
        x1 = copy(x.float())
        x = self.conv1(x, edge_index)
        x2 = copy(x)
        root_extend = torch.zeros(len(batch), x1.size(1)).to(root_index.device)
        batch_size = max(batch) + 1

        for num_batch in range(batch_size):
            index = (torch.eq(batch, num_batch))
            root_extend[index] = x1[root_index[num_batch]]
        x = torch.cat((x, root_extend), 1)

        x = F.relu(x)
        x = F.dropout(x, p=self.dropout_ratio, training=self.training)
        x = self.conv2(x, edge_index)
        x = F.relu(x)
        root_extend = torch.zeros(len(batch), x2.size(1)).to(root_index.device)
        for num_batch in range(batch_size):
            index = (torch.eq(batch, num_batch))
            root_extend[index] = x2[root_index[num_batch]]
        x = torch.cat((x, root_extend), 1)
        out = scatter_mean(x, batch, dim=0)

        return out


class BiGCN(AbstractModel):
    """
    Rumor Detection on Social Media with Bi-Directional Graph Convolutional Networks
    """
    def __init__(self, feature_size: int, hidden_size=128, out_size=128):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): dimension of hidden layer. Default=128
            out_size (int): dimension of output layer. Default=128
        """
        super(BiGCN, self).__init__()
        # top-down
        self.TDRumorGCN = _RumorGCN(feature_size, hidden_size, out_size)
        # bottom-up
        self.BURumorGCN = _RumorGCN(feature_size, hidden_size, out_size)
        self.fc = torch.nn.Linear((out_size + hidden_size) * 2, 2)

    def forward(self, x: Tensor, edge_index: Tensor, bu_edge_index: Tensor,
                batch: Tensor, root_index: Tensor):
        """

        Args:
            x (Tensor): node feature, shape=(num_nodes, feature_size), shape=(num_nodes, feature_size)
            edge_index (Tensor): top-down edge index, shape=(2, num_edges)
            bu_edge_index (Tensor): bottom-up edge index, shape=(2, num_edges)
            batch (Tensor): index of graph each node belongs to, shape=(num_nodes,)
            root_index (Tensor): index of root node for each graph, shape=(num_graphs,)

        Returns:
            out (Tensor): prediction of being fake, shape=(batch_size, 2)
        """
        td_x = self.TDRumorGCN(x, edge_index, batch, root_index)
        bu_x = self.BURumorGCN(x, bu_edge_index, batch, root_index)
        x = torch.cat((td_x, bu_x), 1)
        out = self.fc(x)
        return out

    def calculate_loss(self, data) -> torch.Tensor:
        output = self.forward(data.x, data.edge_index, data.BU_edge_index,
                              data.batch, data.root_index)
        loss_fn = nn.CrossEntropyLoss()
        loss = loss_fn(output, data.y)
        return loss

    def predict(self, data_without_label) -> torch.Tensor:
        output = self.forward(data_without_label.x,
                              data_without_label.edge_index,
                              data_without_label.BU_edge_index,
                              data_without_label.batch,
                              data_without_label.root_index)
        return torch.softmax(output, dim=1)



import pickle
from typing import List, Any, Dict

import yaml
from torch.optim import AdamW
from torch.utils.data import random_split, DataLoader

from faknow.data.dataset.nep_dataset import NEPDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.social_context.nep import NEP
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import dict2str

__all__ = ['run_nep', 'run_nep_from_yaml']


def run_nep(post_simcse, avg_mac, avg_mic, p_mac, p_mic, avg_mic_mic, token, label,
            data_ratio: List[float] = None, batch_size=8, num_epochs=10, lr=5e-4, metrics=None,
            device='cpu', **kwargs):
    dataset = NEPDataset(post_simcse, avg_mac, avg_mic, p_mac, p_mic, avg_mic_mic, token, label)

    # split dataset
    if data_ratio is None:
        data_ratio = [0.7, 0.1, 0.2]

    train_size = int(data_ratio[0] * len(dataset))
    val_size = int(data_ratio[1] * len(dataset))
    test_size = len(dataset) - train_size - val_size
    train_set, val_set, test_set = random_split(dataset, [train_size, val_size, test_size])

    train_loader = DataLoader(train_set, batch_size, shuffle=True)
    val_loader = DataLoader(val_set, batch_size, shuffle=False)
    test_loader = DataLoader(test_set, batch_size, shuffle=False)

    model = NEP(**kwargs)
    optimizer = AdamW(filter(lambda p: p.requires_grad,
                             model.parameters()), lr)
    evaluator = Evaluator(metrics)

    trainer = BaseTrainer(model, evaluator, optimizer, device=device)
    trainer.fit(train_loader, num_epochs, val_loader)

    test_result = trainer.evaluate(test_loader)
    print(f"test result: {dict2str(test_result)}")


def _parse_kargs(config: Dict[str, Any]) -> Dict[str, Any]:
    with open(config.pop('data'), 'rb') as f:
        config.update(pickle.load(f))
    return config


def run_nep_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_nep(**_parse_kargs(_config))
